
#include "SlippyMapTile.h"
#include "gbos1936/Gbos1936.h"
#include "StringUtils.h"
#include "Tile.h"
#include "ImgMagick.h"
#include "ImageWarpByFunc.h"
#include "ganzc/LatLong-OSGBconversion.h"
#include "ReadKmlFile.h"
#include <time.h>

//#include <wand/magick-wand.h>
//#include <wand/drawing-wand.h>
//#include <Magick++.h>
//rsync -r /var/www/os7 timsc@old-dev.openstreetmap.org:/home/timsc/public_html
//rsync -r /var/www/openlayers-os7 timsc@old-dev.openstreetmap.org:/home/timsc/public_html

#include <boost/filesystem.hpp>   // includes all needed Boost.Filesystem declarations
namespace fs = boost::filesystem;

#include <list>
#include <iostream>
#include <exception>
#include <vector>
using namespace std;
//using namespace Magick;

struct RescaleParams
{
public:
	double xscale, yscale, xori, yori;
};

vector<double> RescaleTransform(vector<double> in, void *userPtr)
{
	struct RescaleParams &params = *(struct RescaleParams *)userPtr;
	vector<double> out;
	out.push_back(in[0] * params.xscale - params.xori);
	out.push_back(in[1] * params.yscale - params.yori);
	return out;
}

int GetResizedSubimage(class Tile &src, class Tile &dst, class ImgMagick &imageIn, class ImgMagick &imageOut)
{
	struct RescaleParams params;
	params.xscale = (src.lonmax - src.lonmin) * dst.sx / ((dst.lonmax - dst.lonmin) * src.sx);
	params.yscale = (src.latmax - src.latmin) * dst.sy / ((dst.latmax - dst.latmin) * src.sy);
	params.xori = params.xscale * src.sx * (dst.lonmin - src.lonmin) / (src.lonmax - src.lonmin);
	params.yori = params.yscale * src.sy * (src.latmax - dst.latmax) / (src.latmax - src.latmin);

	class ImageWarpByFunc warp;
	warp.Warp(imageIn, imageOut, RescaleTransform, &params);

	//cout << "scale " << xscale << "\t" << yscale << endl;
	//cout << "ori " << xori << "\t" << yori << endl;

	//double lat, lon, alt;
	//ConvertGbos1936ToWgs84(302000,588000,0.0,lat,lon, alt);
	//cout << "testA " << lat << "," << lon << endl;
	//cout << "test " << (lon - dstW) * dstWidth / (dstE - dstW) << "," << (dstN - lat) * dstHeight / (dstN - dstS) << endl;
	//cout << dst.sx<<","<<dst.sy << endl;
	


	return 1;
}

//************************

class CopyPixelsWithOsMask
{
public:
	int gnorth, gsouth, geast, gwest, boxset;

	CopyPixelsWithOsMask();
	void UpdateBoundingBox(const char *mapref);
	int CheckIfInBox(double lat, double lon);
	void CopyPixels(class ImgMagick &imageIn, class ImgMagick &imageOut, class Tile &tile);
};

CopyPixelsWithOsMask::CopyPixelsWithOsMask()
{
	gnorth = 0;
	gsouth = 0;
	geast = 0;
	gwest = 0;
	boxset = 0;
}

void CopyPixelsWithOsMask::UpdateBoundingBox(const char *mapref)
{
	int dEasting=0, dNorthing=0;
	OSGBGridRefToRefCoords(mapref,dEasting,dNorthing);

	if(!boxset || gsouth > dNorthing) gsouth = dNorthing;
	if(!boxset || gnorth < dNorthing) gnorth = dNorthing;
	if(!boxset || geast < dEasting) geast = dEasting;
	if(!boxset || gwest > dEasting) gwest = dEasting;
	boxset = 1;
	//double lat=-1.0, lon=-1.0, alt = -1.0;
	//ConvertGbos1936ToWgs84(dEasting, dNorthing,0.0, lat, lon, alt);
}

int CopyPixelsWithOsMask::CheckIfInBox(double lat, double lon)
{
	double pnorth,peast,palt;
	ConvertWgs84ToGbos1936(lat,lon, 0.0, peast,pnorth,palt);
	if(pnorth < gsouth) return 0;
	if(pnorth > gnorth) return 0;
	if(peast < gwest) return 0;
	if(peast > geast) return 0;

	return 1;
}

void CopyPixelsWithOsMask::CopyPixels(class ImgMagick &imageIn, class ImgMagick &imageOut, class Tile &tile)
{
	int width = imageIn.GetWidth();
	int height = imageIn.GetHeight();
	int chans = imageIn.GetNumChannels();

	for(int i=0;i<width;i++)
	for(int j=0;j<height;j++)
	for(int k=0;k<chans;k++)
	{
		double lat, lon;
		tile.UnProject(i, j, lat, lon);
		int inMask = this->CheckIfInBox(lat, lon);
		if(!inMask) continue;

		double val = imageIn.GetPix(i,j,k);
		imageOut.SetPix(i,j,k,val);
	}

}

void ImageToBlack(class ImgMagick &image)
{
	int width = image.GetWidth();
	int height = image.GetHeight();
	int chans = image.GetNumChannels();

	for(int i=0;i<width;i++)
	for(int j=0;j<height;j++)
	for(int k=0;k<chans;k++)
	{
		image.SetPix(i,j,k,0);
	}
}

//************************************

class SourceKml
{
public:
	class Tile tile;
	class ImgMagick image;
	vector<string> bounds;
	string imgFilename;
	clock_t lastAccess;

	SourceKml() {lastAccess = 0;}
	SourceKml(const SourceKml &a) {operator=(a);}
	virtual ~SourceKml() {}
	SourceKml& operator=(const SourceKml& a)
	{
		tile = a.tile;
		image = a.image;
		bounds = a.bounds;
		imgFilename = a.imgFilename;
		lastAccess = a.lastAccess;
		return *this;
	}
};

//***************************************

int main(int argc, char ** argv)
{

	
	//Image imageInOut("step.jpg");
	//imageInOut.crop( Geometry(255,255,0,0) );
	//imageInOut.write("test2.png");

	/*class Tile test;
	vector<string> test2;
	string test3;
	int retOpen = ReadKmlFile("/home/tim/Desktop/gis/os7/rect/73.kml", test, test3, test2);

	exit(0);*/
	
	class Tile dst;
	//cout << long2tile(-3.68, zoom) << "," << lat2tile(54.8333,zoom) << endl;
	//cout << long2tile(-3.04, zoom) << "," << lat2tile(55.2446,zoom) << endl;

	vector<class SourceKml> src;
	class SourceKml temp; 

	{src.push_back(temp);
	class SourceKml &last = src[src.size()-1];
	last.tile.latmax = 55.6549;
        last.tile.latmin = 55.2329;
	last.tile.lonmax = -5.22694;
	last.tile.lonmin = -5.8944;
	//cout << last.tile.latmin << "," << last.tile.lonmin << "," << last.tile.latmax << "," << last.tile.lonmax << endl;
	last.imgFilename = "/home/tim/Desktop/gis/os7/rect/65.jpg";
	//int ret = last.image.Open(last.imgFilename.c_str());
	//if(ret < 0){cout << "Filed to open image" << endl;exit(0);}
	last.bounds.push_back("NR550000");
	last.bounds.push_back("NR950450");}

	{src.push_back(temp);
	class SourceKml &last = src[src.size()-1];
	last.tile.latmax = 55.636;
        last.tile.latmin = 55.2197;
	last.tile.lonmax = -4.0457;
	last.tile.lonmin = -4.70144;
	//cout << last.tile.latmin << "," << last.tile.lonmin << "," << last.tile.latmax << "," << last.tile.lonmax << endl;
	last.imgFilename = "/home/tim/Desktop/gis/os7/rect/67.jpg";
	//int ret = last.image.Open(last.imgFilename.c_str());
	//if(ret < 0){cout << "Filed to open image" << endl;exit(0);}
	last.bounds.push_back("NX300950");
	last.bounds.push_back("NS700400");}

	{src.push_back(temp);
	class SourceKml &last = src[src.size()-1];
	last.tile.latmax = 55.4459;
        last.tile.latmin = 55.027;
	last.tile.lonmax = -4.58385;
	last.tile.lonmin = -5.24158;
	//cout << last.tile.latmin << "," << last.tile.lonmin << "," << last.tile.latmax << "," << last.tile.lonmax << endl;
	last.imgFilename = "/home/tim/Desktop/gis/os7/rect/72.jpg";
	//int ret = last.image.Open(last.imgFilename.c_str());
	//if(ret < 0){cout << "Filed to open image" << endl;exit(0);}
	last.bounds.push_back("NW950750");
	last.bounds.push_back("NS350200");}

	{src.push_back(temp);
	class SourceKml &last = src[src.size()-1];
	last.tile.latmax = 55.4459;
        last.tile.latmin = 55.027;
	last.tile.lonmax = -4.58385;
	last.tile.lonmin = -5.24158;
	//cout << last.tile.latmin << "," << last.tile.lonmin << "," << last.tile.latmax << "," << last.tile.lonmax << endl;
	last.imgFilename = "/home/tim/Desktop/gis/os7/rect/72.jpg";
	//int ret = last.image.Open(last.imgFilename.c_str());
	//if(ret < 0){cout << "Filed to open image" << endl;exit(0);}
	last.bounds.push_back("NW950750");
	last.bounds.push_back("NS350200");}

	{src.push_back(temp);
	class SourceKml &last = src[src.size()-1];
	last.tile.latmax = 55.232;
        last.tile.latmin = 54.8158;
	last.tile.lonmax = -4.0252;
	last.tile.lonmin = -4.67399;
	//cout << last.tile.latmin << "," << last.tile.lonmin << "," << last.tile.latmax << "," << last.tile.lonmax << endl;
	last.imgFilename = "/home/tim/Desktop/gis/os7/rect/73.jpg";
	//int ret = last.image.Open(last.imgFilename.c_str());
	//if(ret < 0){cout << "Filed to open image" << endl;exit(0);}
	last.bounds.push_back("NX300500");
	last.bounds.push_back("NX700950");}

	{src.push_back(temp);
	class SourceKml &last = src[src.size()-1];
	last.tile.latmax = 55.2409;
        last.tile.latmin = 54.8278;
	last.tile.lonmax = -3.40279;
	last.tile.lonmin = -4.0457;
	//cout << last.tile.latmin << "," << last.tile.lonmin << "," << last.tile.latmax << "," << last.tile.lonmax << endl;
	last.imgFilename = "/home/tim/Desktop/gis/os7/rect/74.jpg";
	//int ret = last.image.Open(last.imgFilename.c_str());
	//if(ret < 0){cout << "Filed to open image" << endl;exit(0);}
	last.bounds.push_back("NX700500");
	last.bounds.push_back("NY100950");}

	{src.push_back(temp);
	class SourceKml &last = src[src.size()-1];
	last.tile.latmax = 55.2446;
        last.tile.latmin = 54.8333;
	last.tile.lonmax = -3.04477;
	last.tile.lonmin = -3.68423;
	//cout << last.tile.latmin << "," << last.tile.lonmin << "," << last.tile.latmax << "," << last.tile.lonmax << endl;
	last.imgFilename = "/home/tim/Desktop/gis/os7/rect/75.jpg";
	//int ret = last.image.Open(last.imgFilename.c_str());
	//if(ret < 0){cout << "Filed to open image" << endl;exit(0);}
	last.bounds.push_back("NX930950");
	last.bounds.push_back("NY330500");}





	class ImgMagick tile;
	tile.SetNumChannels(3);
	tile.SetWidth(256);
	tile.SetHeight(256);
	class ImgMagick outImg;
	outImg.SetNumChannels(3);
	outImg.SetWidth(256);
	outImg.SetHeight(256);

	//int zoom = 14;
	for(unsigned int zoom=13;zoom<=14;zoom++)
	{
	int srcWtile = long2tile(/*-3.68423*/-6.0, zoom);
	int srcEtile = long2tile(0.0, zoom);
	int srcStile = lat2tile(54.0,zoom);
	int srcNtile = lat2tile(56.0,zoom);
	cout << "tiles covered " << srcStile << "," << srcWtile << "," << srcNtile << "," << srcEtile << endl;

	//int tileLon = 8024;
	//int tileLat = 5168;
	for(int tileLon = srcWtile; tileLon <= srcEtile; tileLon ++)
	for(int tileLat = srcNtile; tileLat <= srcStile; tileLat ++)
	{
	
	ImageToBlack(outImg);

	dst.latmax = tile2lat(tileLat, zoom);
	dst.lonmin = tile2long(tileLon, zoom);
	dst.latmin = tile2lat(tileLat+1, zoom);
	dst.lonmax = tile2long(tileLon+1, zoom);
	dst.sx = outImg.GetWidth();
	dst.sy = outImg.GetHeight();

	//Magick::Image image(src.GetWidth(),src.GetHeight(),"RGB",CharPixel,src.GetInternalDataConst());
	
	//cout << size.width() << "," << size.height() << endl;

	int countSrcTilesUsed = 0;
	for(unsigned int t=0;t<src.size();t++)
	{
	class SourceKml &srcKml = src[t];
	ImageToBlack(tile);

	//Check kml src actually contains useful info for this destination tile
	int noOverlap = 0;
	if(srcKml.tile.lonmax < dst.lonmin) noOverlap = 1;
	if(srcKml.tile.lonmin > dst.lonmax) noOverlap = 1;
	if(srcKml.tile.latmax < dst.latmin) noOverlap = 1;
	if(srcKml.tile.latmin > dst.latmax) noOverlap = 1;
	if(noOverlap) continue;	
	countSrcTilesUsed ++;
	if(countSrcTilesUsed == 1)
	{
		cout << "Generating tile " << tileLon << "," << tileLat << endl;
		cout << "Tile area " << dst.latmin << "," << dst.lonmin << "," << dst.latmax << "," << dst.lonmax << endl;
	}
	cout << "Copying pixels from " << srcKml.imgFilename << endl;

	//Check map is in memory
	if(!srcKml.image.Ready())
	{
		//Count how many tiles in memory
		int count = 0;
		for(unsigned int i=0;i<src.size();i++)
			if(src[i].image.Ready()) count ++;
		cout << "Currently " << count << " tiles in memory" << endl;
		
		int maxTilesLoaded = 4;
		clock_t lowTime = 0; int lowTimeSet = 0; int lowIndex = -1;
		if(count >= maxTilesLoaded)
		{
			//Need to unload a tile. Find tile that has not been recently used.
			for(unsigned int i=0;i<src.size();i++)
			{
				//cout << "accessed " << src[i].imgFilename << " at " << src[i].lastAccess << endl;
				if((src[i].lastAccess < lowTime || lowTimeSet == 0) && src[i].image.Ready())
				{
					lowTime = src[i].lastAccess;
					lowIndex = i;
					//cout << i << "," << src[i].lastAccess << "," << lowTime << "," << src[i].lastAccess < lowTime << endl;
					//cout << "lowIndex " << lowIndex << endl;
				}
				lowTimeSet = 1;
			}
			
			//Unload tile
			if(lowIndex != -1)
			{
				cout << "Unloading " << src[lowIndex].imgFilename << endl;
				src[lowIndex].image.Clear();
			}		
		}
		
		//Load required tile into mem
		cout << "Loading " << srcKml.imgFilename << endl;	
		int ret = srcKml.image.Open(srcKml.imgFilename.c_str());
		if(ret < 0) {cout << "Failed to open image " << srcKml.imgFilename << endl;exit(0);}
		srcKml.tile.sx = srcKml.image.GetWidth();
		srcKml.tile.sy = srcKml.image.GetHeight();
	}

	//Update access timer for this tile
	srcKml.lastAccess = clock();

	//Get location in source image containing area of interest
	double xmin = (double)srcKml.tile.sx * (dst.lonmin - srcKml.tile.lonmin) / (srcKml.tile.lonmax - srcKml.tile.lonmin);
	double xmax = (double)srcKml.tile.sx * (dst.lonmax - srcKml.tile.lonmin) / (srcKml.tile.lonmax - srcKml.tile.lonmin);
	double ymin = (double)srcKml.tile.sy * (srcKml.tile.latmax - dst.latmax) / (srcKml.tile.latmax - srcKml.tile.latmin);	
	double ymax = (double)srcKml.tile.sy * (srcKml.tile.latmax - dst.latmin) / (srcKml.tile.latmax - srcKml.tile.latmin);	
	xmin = (int)(xmin - 100);
	xmax = (int)(xmax + 100);
	ymin = (int)(ymin - 100);
	ymax = (int)(ymax + 100);
	if(xmin < 0) xmin = 0;
	if(ymin < 0) ymin = 0;
	if(xmax > srcKml.tile.sx) xmax = srcKml.tile.sx;
	if(ymax > srcKml.tile.sy) ymax = srcKml.tile.sy;

	//cout << "src image roi " << xmin << "," <<xmax<< "\t" << ymin << "," << ymax << endl;
	/*class Tile inter;
	inter.lonmin = ((double)xmin * (src.lonmax - src.lonmin) / (double)src.sx) + src.lonmin;
	inter.lonmax = ((double)xmax * (src.lonmax - src.lonmin) / (double)src.sx) + src.lonmin;
	inter.latmin = src.latmax - ((double)ymax * (src.latmax - src.latmin) / (double)src.sy);
	inter.latmax = src.latmax - ((double)ymin * (src.latmax - src.latmin) / (double)src.sy);
	inter.sx = xmax - xmin;
	inter.sy = ymax - ymin;*/
	//cout << inter.sx << "," << inter.sy << endl;
	//cout << "inter tile area " << inter.lonmin << "," << inter.lonmax << "\t" << inter.latmin << "," <<inter.latmax << endl;



	//cout << "a" << endl;
	//GetSubimage(image,src,temp,inter);
	//temp.write("temp.png");

	//cout << "b" << endl;
	GetResizedSubimage(srcKml.tile,dst,srcKml.image,tile);
	//GetResizedSubimage(inter,dst,temp);
	//cout << "c" << endl;
	
	class CopyPixelsWithOsMask mask;
	for(unsigned int i=0;i<srcKml.bounds.size();i++)
		mask.UpdateBoundingBox(srcKml.bounds[i].c_str());
	mask.CopyPixels(tile, outImg, dst);

	}

	//Don't write blank tiles
	if(countSrcTilesUsed == 0) 
		continue;

	//outImg = tile;

	string outFilename;
	outFilename += IntToString(zoom);

	if ( !fs::exists( outFilename ) )
	{
		fs::create_directory( outFilename );
	}

	outFilename += "/";
	outFilename += IntToString(tileLon);

	if ( !fs::exists( outFilename ) )
	{
		fs::create_directory( outFilename );
	}

	outFilename += "/";
	outFilename += IntToString(tileLat);
	outFilename += ".jpg";

	   	outImg.Save(outFilename.c_str());
		//temp.syncPixels();
		//image.syncPixels();

	
	} //End of tile loop
	} //End of zoom loop
}

